import datetime

from rollbot import RollbotFailure, as_plugin, pop_arg, as_group_singleton


@as_group_singleton
class DnDSessionV2:
    session_time: datetime.datetime


WEEKDAYS = {
    "monday": 0,
    "tuesday": 1,
    "wednesday": 2,
    "thursday": 3,
    "friday": 4,
    "saturday": 5,
    "sunday": 6
}


def parse_datetime(arg_text):
    if arg_text is None:
        return

    day_date, arg_text = pop_arg(arg_text)
    day_date = day_date.lower()
    if day_date in WEEKDAYS:
        today = datetime.date.today()
        day_diff = WEEKDAYS[day_date] - today.weekday()
        next_day = today + datetime.timedelta((day_diff) % 7)
        month = next_day.month
        day = next_day.day
        year = next_day.year
    else:
        date_parts = day_date.split("/")
        if len(date_parts) == 3:
            month = date_parts[0]
            day = date_parts[1]
            year = date_parts[2]
        elif len(date_parts) == 2:
            month = date_parts[0]
            day = date_parts[1]
            year = datetime.date.today().year
        else:
            return  # invalid format

    hour = 19
    minute = 45

    if arg_text is None:
        return datetime.datetime(
            int(year), int(month), int(day),
            int(hour), int(minute)
        )

    time, arg_text = pop_arg(arg_text)
    time_parts = time.split(":")
    hour = time_parts[0]
    if len(time_parts) > 1:
        minute = time_parts[1]

    ampm, arg_text = pop_arg(arg_text)

    if ampm is None:
        return datetime.datetime(
            int(year), int(month), int(day),
            int(hour), int(minute)
        )

    if ampm.lower() == "pm":
        hour = int(hour) + 12

    return datetime.datetime(
        int(year), int(month), int(day),
        int(hour), int(minute)
    )


def get_fmt_next(nxt):
    if nxt is None or nxt.session_time is None or nxt.session_time < datetime.datetime.now():
        return "No upcoming session found."
    return "Next DnD session scheduled for " + nxt.session_time.strftime("%b %d, %Y, at %I:%M %p")


def session_next(session_data, arg_text):
    dt = parse_datetime(arg_text)
    if dt is None:
        return RollbotFailure.INVALID_ARGUMENTS.with_reason("Could not process argument as a date/time.")
    if dt < datetime.datetime.now():
        return RollbotFailure.INVALID_ARGUMENTS.with_reason("Cannot schedule sessions in the past.")
    
    session_data.session_time = dt
    return get_fmt_next(session_data)


def session_late(session_data, arg_text):
    if session_data.session_time is None:
        return "No upcoming session found."

    now = datetime.datetime.now()
    nxt = session_data.session_time
    if nxt >= now:
        return "You're not late yet!"

    late = (now - nxt).total_seconds() / 60
    return f"You're running {late:.2f} minutes late..."


def session_cancel(session_data, arg_text):
    if session_data.session_time is None:
        return "No upcoming session found."
    
    session_data.session_time = None
    return "Sigh, session cancelled..."


SUBCOMMANDS = dict(
    next=session_next,
    view=lambda session_data, _: get_fmt_next(session_data),
    late=session_late,
    cancel=session_cancel
)

NEEDS_ADMIN = ("next", "cancel")

VALID_TEXT = "Valid subcommands are: " + ", ".join(f"!{sc}" for sc in SUBCOMMANDS)


@as_plugin
def session(bot, subc, session_data: DnDSessionV2):
    if subc is None:
        return RollbotFailure.MISSING_SUBCOMMAND.with_reason(VALID_TEXT)

    if subc.command not in SUBCOMMANDS:
        return RollbotFailure.INVALID_SUBCOMMAND.with_reason(VALID_TEXT)

    if subc.command in NEEDS_ADMIN and not subc.from_admin:
        return RollbotFailure.PERMISSIONS

    return SUBCOMMANDS[subc.command](session_data, subc.raw_args)

    